"""Preset-based config tooling for BEACH."""

from .core import (
    CASE_FILENAME,
    DEFAULT_PRESET_NAMES,
    RENDERED_FILENAME,
    CaseDocument,
    CaseSpecError,
    ListedPreset,
    ConfigError,
    MergeError,
    PresetResolutionError,
    RenderResult,
    RenderValidationError,
    beachx_home,
    builtin_presets_dir,
    default_case_document,
    default_preset_fragment,
    extract_preset_fragment,
    list_saved_cases,
    list_presets,
    load_case_document,
    load_saved_case_path,
    merge_fragments,
    normalize_preset_section,
    normalize_preset_name,
    preset_category,
    preset_target_path,
    project_presets_dir,
    render_beach_toml,
    render_case_document,
    render_case_file,
    render_case_toml,
    render_preset_toml,
    resolve_preset,
    save_case_file,
    save_preset_fragment,
    saved_cases_dir,
    semantic_diff,
    user_presets_dir,
    validate_rendered_config,
)

__all__ = [
    "CASE_FILENAME",
    "DEFAULT_PRESET_NAMES",
    "RENDERED_FILENAME",
    "CaseDocument",
    "CaseSpecError",
    "ListedPreset",
    "ConfigError",
    "MergeError",
    "PresetResolutionError",
    "RenderResult",
    "RenderValidationError",
    "beachx_home",
    "builtin_presets_dir",
    "default_case_document",
    "default_preset_fragment",
    "extract_preset_fragment",
    "list_saved_cases",
    "list_presets",
    "load_case_document",
    "load_saved_case_path",
    "merge_fragments",
    "normalize_preset_section",
    "normalize_preset_name",
    "preset_category",
    "preset_target_path",
    "project_presets_dir",
    "render_beach_toml",
    "render_case_document",
    "render_case_file",
    "render_case_toml",
    "render_preset_toml",
    "resolve_preset",
    "save_case_file",
    "save_preset_fragment",
    "saved_cases_dir",
    "semantic_diff",
    "user_presets_dir",
    "validate_rendered_config",
]
